@@ -15,6 +15,7 @@ class Line(Visible):
     "Base class for all objects which connect points in Feynman diagrams"
     
     def __init__(self, point1, point2):
+        """Constructor."""
         self.p1 = point1
         self.p2 = point2
         self.styles = []
@@ -27,7 +28,7 @@ class Line(Visible):
         FeynDiagram.currentDiagram.add(self)
 
 
-    def addLabel(self, text, pos=0.5, displace=-0.25, angle = 0):
+    def addLabel(self, text, pos=0.5, displace=-0.25, angle = 0, size=pyx.text.size.normalsize):
         """Add a LaTeX label to this line, either via parameters or actually as
         a TeXLable object."""
         if config.getOptions().DEBUG:
@@ -35,17 +36,20 @@ class Line(Visible):
         #if text.__class__ == "Label":
         #    self.labels.append(label)
         #else:
-        self.labels.append(LineLabel(text=text, line=self, pos=pos, displace=displace, angle=angle))
+        self.labels.append(LineLabel(text=text, line=self, pos=pos, displace=displace, angle=angle, size=size))
         if config.getOptions().DEBUG:
             print "Labels = " + str(self.labels)
         return self
 
     def addParallelArrow(self, pos=0.5, displace=0.3, length=0.5*pyx.unit.v_cm,
-                 size=6*pyx.unit.v_pt, angle=45, constriction=0.8, sense=+1):
+                 size=6*pyx.unit.v_pt, angle=45, constriction=0.8, sense=+1,
+                 curved=False,stems=1,stemsep=0.03):
+        """Add an arrow pointing along the line."""
         self.labels.append(ParallelArrow(self, pos=pos, displace=displace,
                                          length=length, size=size, angle=angle,
-                                         constriction=constriction,
-                                         sense=sense))
+                                         constriction=constriction,sense=sense,
+                                         curved=curved,stems=stems,
+                                         stemsep=stemsep))
         return self
 
             
@@ -126,14 +130,14 @@ class Line(Visible):
             ny *= -1
         arcpoint = Point(middle.x() + amount * nx, middle.y() + amount * ny)
         if config.getOptions().VDEBUG:
-            FeynDiagram.currentCanvas.stroke(
+            FeynDiagram.currenDiagram.currentCanvas.stroke(
                 pyx.path.line(middle.x(), middle.y(), arcpoint.x(),
                               arcpoint.y()), [color.rgb.blue] )
         self.arcThru(arcpoint)
         if config.getOptions().DEBUG:
             print self.getVisiblePath()
         if config.getOptions().VDEBUG:
-            FeynDiagram.currentCanvas.stroke(self.getVisiblePath(), [color.rgb.blue])
+            FeynDiagram.currenDiagram.currentCanvas.stroke(self.getVisiblePath(), [color.rgb.blue])
         return self
 
 
@@ -184,9 +188,9 @@ class Line(Visible):
             circle = pyx.path.circle(*cargs)
             line = pyx.path.line( self.p1.x(), self.p1.y(), arccenter.x(), arccenter.y())
             if config.getOptions().VDEBUG:
-                FeynDiagram.currentCanvas.stroke(line, [color.rgb.green])
-            as, bs = circle.intersect(line)
-            subpaths = circle.split(as[0])
+                FeynDiagram.currenDiagram.currentCanvas.stroke(line, [color.rgb.green])
+            ass, bs = circle.intersect(line)
+            subpaths = circle.split(ass[0])
             cpath = subpaths[0]
             return cpath
 
@@ -270,9 +274,9 @@ class Line(Visible):
         p2path = self.p2.getPath()
         vispath = self.getPath()
         if config.getOptions().VDEBUG:
-            FeynDiagram.currentCanvas.stroke(vispath, [color.rgb.green])
+            FeynDiagram.currenDiagram.currentCanvas.stroke(vispath, [color.rgb.green])
         if p1path:
-            as, bs = p1path.intersect(vispath)
+            ass, bs = p1path.intersect(vispath)
             for b in bs:
                 subpaths = vispath.split(b)
                 if len(subpaths) > 1:
@@ -284,14 +288,14 @@ class Line(Visible):
                             math.fabs(pyx.unit.tocm(x.arclen() - y.arclen()))) )
                     vispath = subpaths[-1]
                     if config.getOptions().VDEBUG:
-                        FeynDiagram.currentCanvas.stroke(subpaths[0], [color.rgb.blue])
+                        FeynDiagram.currenDiagram.currentCanvas.stroke(subpaths[0], [color.rgb.blue])
                 if config.getOptions().VDEBUG:
-                    for a in as:
+                    for a in ass:
                         ix, iy = p1path.at(a)
-                        FeynDiagram.currentCanvas.fill(pyx.path.circle(ix, iy, 0.05),
+                        FeynDiagram.currenDiagram.currentCanvas.fill(pyx.path.circle(ix, iy, 0.05),
                                                        [color.rgb.green])
         if p2path: 
-            as, bs = p2path.intersect(vispath)
+            ass, bs = p2path.intersect(vispath)
             for b in bs:
                 subpaths = vispath.split(b)
                 if len(subpaths) > 1:
@@ -303,14 +307,14 @@ class Line(Visible):
                             math.fabs(pyx.unit.tocm(x.arclen() - y.arclen()))) )
                     vispath = subpaths[-1]
                     if config.getOptions().VDEBUG:
-                        FeynDiagram.currentCanvas.stroke(subpaths[0], [color.rgb.red])
+                        FeynDiagram.currenDiagram.currentCanvas.stroke(subpaths[0], [color.rgb.red])
                 if config.getOptions().VDEBUG:
-                    for a in as:
+                    for a in ass:
                         ix, iy = p2path.at(a)
-                        FeynDiagram.currentCanvas.fill(pyx.path.circle(ix, iy, 0.05),
+                        FeynDiagram.currenDiagram.currentCanvas.fill(pyx.path.circle(ix, iy, 0.05),
                                                        [color.rgb.blue])
         if config.getOptions().VDEBUG:
-            FeynDiagram.currentCanvas.stroke(vispath, [color.rgb.red])
+            FeynDiagram.currenDiagram.currentCanvas.stroke(vispath, [color.rgb.red])
         #return pyx.path.circle(-2,-1,0.2)
         return vispath
 
@@ -334,6 +338,7 @@ Fermion = Line
 class MultiLine(Line):
     """A class for drawing multiple parallel straight lines."""
     def __init__(self, point1, point2, n=5, dist=0.2):
+        """Constructor."""
         self.p1 = point1
         self.p2 = point2
         self.styles = []
@@ -388,6 +393,23 @@ Sfermion = Scalar
 
 
 
+class Ghost(Line):
+    """A dotted scalar particle line, like a Yang-Mills ghost particle."""
+
+    def draw(self, canvas):
+        """Draw this scalar line on the given canvas."""
+        path = self.getVisiblePath()
+        styles = self.styles + [pyx.style.linestyle.dotted] + self.arrows
+        ## TODO: call base class method?
+        if config.getOptions().DEBUG:
+            print "Drawing " + str(self.__class__) + " with styles = " + str(styles)
+            print path
+        canvas.stroke(path, styles)
+        for l in self.labels:
+            l.draw(canvas)
+
+
+
 ## DecoratedLine base class
 class DecoratedLine(Line):
     """Base class for spring and sine-like lines"""
@@ -409,6 +431,7 @@ class DecoratedLine(Line):
 class Gluon(DecoratedLine):
     """A line with a cycloid deformation"""
     def __init__(self, point1, point2):
+        """Constructor."""
         self.p1 = point1
         self.p2 = point2
         self.styles = []
@@ -428,6 +451,7 @@ class Gluon(DecoratedLine):
 
 
     def set3D(self, is3D=True, skipsize=pyx.unit.length(0.04), parity=0):
+        """Make this line display in '3D'."""
         self.is3D = is3D
         self.skipsize3D = skipsize
         self.parity3D = parity
@@ -497,7 +521,7 @@ class Gluon(DecoratedLine):
             canvas.stroke(mypath, styles)
         else:
             para = pyx.deformer.parallel(0.001)
-            as, bs, cs = para.normpath_selfintersections(mypath.normpath(), epsilon=0.01)
+            ass, bs, cs = para.normpath_selfintersections(mypath.normpath(), epsilon=0.01)
             coil_params = []
             for b in bs:
                 coil_params.append(b[self.parity3D] - self.skipsize3D)
@@ -518,6 +542,7 @@ class Gluon(DecoratedLine):
 class Vector(DecoratedLine):
     """A line with a sinoid deformation"""
     def __init__(self, point1, point2, amplitude=0.25, frequency=1.0):
+        """Constructor."""
         self.p1 = point1
         self.p2 = point2
         self.styles = []
@@ -598,6 +623,7 @@ Photon = Vector
 class Graviton(DecoratedLine):
     """A line with a double sinoid deformation"""
     def __init__(self, point1, point2):
+        """Constructor."""
         self.p1 = point1
         self.p2 = point2
         self.styles = []
@@ -615,6 +641,7 @@ class Graviton(DecoratedLine):
 
 
     def set3D(self, is3D=True, skipsize=pyx.unit.length(0.04), parity=0):
+        """Make this line display in '3D'."""
         self.is3D = is3D
         self.skipsize3D = skipsize
         self.parity3D = parity
@@ -670,13 +697,13 @@ class Graviton(DecoratedLine):
             canvas.stroke(mypath1, styles)
             canvas.stroke(mypath2, styles)
         else:
-            as, bs = mypath1.intersect(mypath2)
+            ass, bs = mypath1.intersect(mypath2)
             params1, params2 = [], []
 
             parity1 = True
             if self.parity3D == 0:
                 parity1 = False
-            for a in as[1:]: ## TODO: better endpoint cut vetoing
+            for a in ass[1:]: ## TODO: better endpoint cut vetoing
                 if parity1:
                     params1.append(a - self.skipsize3D)
                     params1.append(a + self.skipsize3D)
@@ -703,7 +730,6 @@ class Graviton(DecoratedLine):
                     canvas.stroke(pathbit, styles)
                 on = not on
 
-
         for l in self.labels:
             l.draw(canvas)
 
@@ -713,6 +739,7 @@ class Gaugino(DecoratedLine):
     """A line with a sinoid deformation and a normal line"""
     
     def __init__(self, point1, point2):
+        """Constructor."""
         self.p1 = point1
         self.p2 = point2
         self.styles = []
@@ -786,13 +813,13 @@ class Gaugino(DecoratedLine):
             canvas.stroke(mypath1, styles)
             canvas.stroke(mypath2, styles)
         else:
-            as, bs = mypath1.intersect(mypath2)
+            ass, bs = mypath1.intersect(mypath2)
             params1, params2 = [], []
 
             parity1 = True
             if self.parity3D == 0:
                 parity1 = False
-            for a in as:
+            for a in ass:
                 if parity1:
                     params1.append(a - self.skipsize3D)
                     params1.append(a + self.skipsize3D)
@@ -826,6 +853,7 @@ class Gaugino(DecoratedLine):
 class Gluino(DecoratedLine):
     """A line with a cycloid deformation and a normal line"""
     def __init__(self, point1, point2):
+        """Constructor."""
         self.p1 = point1
         self.p2 = point2
         self.styles = []
@@ -843,6 +871,7 @@ class Gluino(DecoratedLine):
 
 
     def set3D(self, is3D=True, skipsize=pyx.unit.length(0.04), parity=0):
+        """Make this line display in '3D'."""
         self.is3D = is3D
         self.skipsize3D = skipsize
         self.parity3D = parity
@@ -908,13 +937,13 @@ class Gluino(DecoratedLine):
             canvas.stroke(mypath1, styles)
             canvas.stroke(mypath2, styles)
         else:
-            as, bs = mypath1.intersect(mypath2)
+            ass, bs = mypath1.intersect(mypath2)
             params1, params2 = [], []
 
             parity1 = True
             if self.parity3D == 0:
                 parity1 = False
-            for a in as:
+            for a in ass:
                 if parity1:
                     params1.append(a - self.skipsize3D)
                     params1.append(a + self.skipsize3D)
@@ -954,13 +983,126 @@ class Gluino(DecoratedLine):
             l.draw(canvas)
 
 
+class Gravitino(DecoratedLine):
+    """A line with a double sinoid deformation and a simple line"""
+    def __init__(self, point1, point2):
+        """Constructor."""
+        self.p1 = point1
+        self.p2 = point2
+        self.styles = []
+        self.arcthrupoint = None
+        self.is3D = False
+        self.skipsize3D = pyx.unit.length(0.04)
+        self.parity3D = 0
+        self.inverted = False
+        self.arrows = []
+        self.labels = []
+        self.arcradius = pyx.unit.length(0.25)
+        self.linetype = "gravitino"
+        ## Add this to the current diagram automatically
+        FeynDiagram.currentDiagram.add(self)
 
 
+    def set3D(self, is3D=True, skipsize=pyx.unit.length(0.04), parity=0):
+        """Make this line display in '3D'."""
+        self.is3D = is3D
+        self.skipsize3D = skipsize
+        self.parity3D = parity
+        return self
+
+
+    def invert(self):
+        """Reflect the decoration in the line itself."""
+        self.inverted = not self.inverted
+        return self
+
+
+    def getDeformedPath(self, sign = 1):
+        """Get the path with the decorative deformation."""
+        intwindings = int(0.6 * pyx.unit.tocm(self.getVisiblePath().arclen()) /
+                          pyx.unit.tocm(self.arcradius))
+
+        vispath = self.getVisiblePath()
+        curveradii = vispath.curveradius([i/10.0 for i in range(0,11)])
+        mincurveradius = None
+        for curveradius in curveradii:
+            try:
+                curveradius = abs(curveradius/pyx.unit.m)
+                #if config.getOptions().DEBUG:
+                #    print self.__class__, "- curve radius = ", curveradius
+                if (mincurveradius is None or curveradius < mincurveradius):
+                    mincurveradius = curveradius
+            except: 
+                pass
+        numhloopcurves = 5
+        if mincurveradius is not None:
+            numhloopcurves += int(0.1/mincurveradius)
+        if config.getOptions().DEBUG:
+            print self.__class__, "- min curvature radius = ", mincurveradius, "->", numhloopcurves, "curves/hloop"                    
+
+        defo = pyx.deformer.cycloid(self.arcradius, intwindings, curvesperhloop=numhloopcurves,
+                                    skipfirst=0.0, skiplast=0.0, turnangle=0, sign=sign)
+        return defo.deform(vispath)
+
+        
+    def draw(self, canvas):
+        """Draw the line on the supplied canvas."""
+        styles = self.styles + self.arrows
+        if config.getOptions().DEBUG:
+            print "Drawing " + str(self.__class__) + " with styles = " + str(styles)
+        mypath = [self.getVisiblePath(), self.getDeformedPath(+1), self.getDeformedPath(-1)]
+        if self.inverted:
+            mypath = mypath[::-1]
+        if config.getOptions().DRAFT or not self.is3D:
+            for i in range(3):
+                canvas.stroke(mypath[i], styles)
+        else:
+            ass, bs = mypath[0].intersect(mypath[1])
+            ass, cs = mypath[0].intersect(mypath[2])
+            ps = [ass[:], bs[:], cs[:]]
+
+            for i in range(3):
+                params = []
+                for a in range(len(ps[i])): ## TODO: better endpoint cut vetoing
+                    if (a%3)!=i:
+                        params.append(ps[i][a] - self.skipsize3D)
+                        params.append(ps[i][a] + self.skipsize3D)
+                pathbits = mypath[i].split(params)
+                on = True
+                for pathbit in pathbits:
+                    if on:
+                        canvas.stroke(pathbit, styles)
+                    on = not on
+
+        for l in self.labels:
+            l.draw(canvas)
+
+
+
+class Phantom(DecoratedLine):
+    """An invisible line."""
+    def __init__(self,  point1, point2):
+        """Constructor."""
+        self.p1 = point1
+        self.p2 = point2
+        self.linetype = "phantom"
+
+    def draw(self, canvas):
+        """Draw the line on the supplied canvas (does nothing for a phantom)."""
+        return
+
 # A dictionary for mapping FeynML line types to line classes
 NamedLine = { "higgs"    : Higgs,
               "photon"   : Photon,
+              "vector"   : Photon,
               "gluon"    : Gluon,
               "fermion"  : Fermion,
               "graviton" : Graviton,
               "gaugino"  : Gaugino,
-              "gluino"   : Gluino }
+              "gluino"   : Gluino,
+              "gravitino": Gravitino,
+              "scalar"   : Higgs,
+              "ghost"    : Ghost,
+              "phantom"  : Phantom
+             }
+
